# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings
from unittest.mock import MagicMock

import matplotlib as mpl
import numpy as np
import pytest
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.contour import QuadContourSet
from matplotlib.figure import Figure
from numpy.testing import assert_allclose
from packaging.version import Version

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.visualization.wcsaxes.core import WCSAxes
from astropy.visualization.wcsaxes.frame import (
    EllipticalFrame,
    RectangularFrame,
    RectangularFrame1D,
)
from astropy.visualization.wcsaxes.ticklabels import TickLabels
from astropy.visualization.wcsaxes.transforms import CurvedTransform
from astropy.visualization.wcsaxes.utils import get_coord_meta
from astropy.wcs import WCS
from astropy.wcs.wcsapi import HighLevelWCSWrapper, SlicedLowLevelWCS

ft_version = Version(mpl.ft2font.__freetype_version__)
FREETYPE_261 = ft_version == Version("2.6.1")

# We cannot use matplotlib.checkdep_usetex() anymore, see
# https://github.com/matplotlib/matplotlib/issues/23244
TEX_UNAVAILABLE = True


def test_grid_regression(ignore_matplotlibrc):
    # Regression test for a bug that meant that if the rc parameter
    # axes.grid was set to True, WCSAxes would crash upon initialization.
    mpl.rc("axes", grid=True)
    fig = Figure(figsize=(3, 3))
    WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])


def test_format_coord_regression(ignore_matplotlibrc, tmp_path):
    # Regression test for a bug that meant that if format_coord was called by
    # Matplotlib before the axes were drawn, an error occurred.
    fig = Figure(figsize=(3, 3))
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    fig.add_axes(ax)
    assert ax.format_coord(10, 10) == ""
    assert ax.coords[0].format_coord(10) == ""
    assert ax.coords[1].format_coord(10) == ""
    fig.savefig(tmp_path / "nothing")
    assert ax.format_coord(10, 10) == "10.0 10.0 (world)"
    assert ax.coords[0].format_coord(10) == "10.0"
    assert ax.coords[1].format_coord(10) == "10.0"


TARGET_HEADER = fits.Header.fromstring(
    """
NAXIS   =                    2
NAXIS1  =                  200
NAXIS2  =                  100
CTYPE1  = 'RA---MOL'
CRPIX1  =                  500
CRVAL1  =                180.0
CDELT1  =                 -0.4
CUNIT1  = 'deg     '
CTYPE2  = 'DEC--MOL'
CRPIX2  =                  400
CRVAL2  =                  0.0
CDELT2  =                  0.4
CUNIT2  = 'deg     '
COORDSYS= 'icrs    '
""",
    sep="\n",
)


@pytest.mark.parametrize("grid_type", ["lines", "contours"])
def test_no_numpy_warnings(ignore_matplotlibrc, tmp_path, grid_type):
    fig = Figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS(TARGET_HEADER))
    ax.imshow(np.zeros((100, 200)))
    ax.coords.grid(color="white", grid_type=grid_type)

    # There should be no warnings raised if some pixels are outside WCS
    # (since this is normal).
    # BUT our own catch_warning was ignoring some warnings before, so now we
    # have to catch it. Otherwise, the pytest filterwarnings=error
    # setting in pyproject.toml will fail this test.
    # There are actually multiple warnings but they are all similar.
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=r".*converting a masked element to nan.*"
        )
        warnings.filterwarnings(
            "ignore", message=r".*No contour levels were found within the data range.*"
        )
        warnings.filterwarnings(
            "ignore", message=r".*np\.asscalar\(a\) is deprecated since NumPy v1\.16.*"
        )
        warnings.filterwarnings(
            "ignore", message=r".*PY_SSIZE_T_CLEAN will be required.*"
        )
        fig.savefig(tmp_path / "test.png")


def test_invalid_frame_overlay(ignore_matplotlibrc):
    # Make sure a nice error is returned if a frame doesn't exist
    fig = Figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS(TARGET_HEADER))
    with pytest.raises(ValueError, match=r"Frame banana not found"):
        ax.get_coords_overlay("banana")

    with pytest.raises(ValueError, match=r"Unknown frame: banana"):
        get_coord_meta("banana")


def test_plot_coord_transform(ignore_matplotlibrc):
    twoMASS_k_header = get_pkg_data_filename("data/2MASS_k_header")
    twoMASS_k_header = fits.Header.fromtextfile(twoMASS_k_header)
    fig = Figure(figsize=(6, 6))
    ax = fig.add_axes(
        [0.15, 0.15, 0.8, 0.8], projection=WCS(twoMASS_k_header), aspect="equal"
    )
    ax.set_xlim(-0.5, 720.5)
    ax.set_ylim(-0.5, 720.5)

    c = SkyCoord(359.76045223 * u.deg, 0.26876217 * u.deg)
    with pytest.raises(TypeError):
        ax.plot_coord(c, "o", transform=ax.get_transform("galactic"))


def test_scatter_coord_transform(ignore_matplotlibrc):
    twoMASS_k_header = get_pkg_data_filename("data/2MASS_k_header")
    twoMASS_k_header = fits.Header.fromtextfile(twoMASS_k_header)
    fig = Figure(figsize=(6, 6))
    ax = fig.add_axes(
        [0.15, 0.15, 0.8, 0.8], projection=WCS(twoMASS_k_header), aspect="equal"
    )
    ax.set_xlim(-0.5, 720.5)
    ax.set_ylim(-0.5, 720.5)

    c = SkyCoord(359.76045223 * u.deg, 0.26876217 * u.deg)
    with pytest.raises(TypeError):
        ax.scatter_coord(c, marker="o", transform=ax.get_transform("galactic"))


def test_set_label_properties(ignore_matplotlibrc):
    # Regression test to make sure that arguments passed to
    # set_xlabel/set_ylabel are passed to the underlying coordinate helpers

    fig = Figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS(TARGET_HEADER))

    ax.set_xlabel("Test x label", labelpad=2, color="red")
    ax.set_ylabel("Test y label", labelpad=3, color="green")

    assert ax.coords[0]._axislabels.get_text() == "Test x label"
    assert ax.coords[0]._axislabels.get_minpad("b") == 2
    assert ax.coords[0]._axislabels.get_color() == "red"

    assert ax.coords[1]._axislabels.get_text() == "Test y label"
    assert ax.coords[1]._axislabels.get_minpad("l") == 3
    assert ax.coords[1]._axislabels.get_color() == "green"

    assert ax.get_xlabel() == "Test x label"
    assert ax.get_ylabel() == "Test y label"


GAL_HEADER = fits.Header.fromstring(
    """
SIMPLE  =                    T / conforms to FITS standard
BITPIX  =                  -32 / array data type
NAXIS   =                    3 / number of array dimensions
NAXIS1  =                   31
NAXIS2  =                 2881
NAXIS3  =                  480
EXTEND  =                    T
CTYPE1  = 'DISTMOD '
CRVAL1  =                  3.5
CDELT1  =                  0.5
CRPIX1  =                  1.0
CTYPE2  = 'GLON-CAR'
CRVAL2  =                180.0
CDELT2  =               -0.125
CRPIX2  =                  1.0
CTYPE3  = 'GLAT-CAR'
CRVAL3  =                  0.0
CDELT3  =                0.125
CRPIX3  =                241.0
""",
    sep="\n",
)


def test_slicing_warnings(ignore_matplotlibrc, tmp_path):
    # Regression test to make sure that no warnings are emitted by the tick
    # locator for the sliced axis when slicing a cube.

    # Scalar case

    wcs3d = WCS(naxis=3)
    wcs3d.wcs.ctype = ["x", "y", "z"]
    wcs3d.wcs.cunit = ["deg", "deg", "km/s"]
    wcs3d.wcs.crpix = [614.5, 856.5, 333]
    wcs3d.wcs.cdelt = [6.25, 6.25, 23]
    wcs3d.wcs.crval = [0.0, 0.0, 1.0]

    fig = Figure()
    with warnings.catch_warnings():
        # https://github.com/astropy/astropy/issues/9690
        warnings.filterwarnings("ignore", message=r".*PY_SSIZE_T_CLEAN.*")
        fig.add_subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))
        fig.savefig(tmp_path / "test.png")

    # Angle case

    wcs3d = WCS(GAL_HEADER)

    with warnings.catch_warnings():
        # https://github.com/astropy/astropy/issues/9690
        warnings.filterwarnings("ignore", message=r".*PY_SSIZE_T_CLEAN.*")
        fig.clear()
        fig.add_subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 2))
        fig.savefig(tmp_path / "test.png")


@pytest.fixture
def tmp_plt_figure():
    import matplotlib.pyplot as plt

    figure = plt.figure()
    yield figure
    plt.close(figure)


@pytest.mark.usefixtures("tmp_plt_figure")
def test_plt_xlabel_ylabel(tmp_path):
    # Regression test for a bug that happened when using plt.xlabel
    # and plt.ylabel with Matplotlib 3.0
    import matplotlib.pyplot as plt

    plt.subplot(projection=WCS())
    plt.xlabel("Galactic Longitude")
    plt.ylabel("Galactic Latitude")
    plt.savefig(tmp_path / "test.png")


def test_grid_type_contours_transform(tmp_path):
    # Regression test for a bug that caused grid_type='contours' to not work
    # with custom transforms

    class CustomTransform(CurvedTransform):
        # We deliberately don't define the inverse, and has_inverse should
        # default to False.

        def transform(self, values):
            return values * 1.3

    transform = CustomTransform()
    coord_meta = {
        "type": ("scalar", "scalar"),
        "unit": (u.m, u.s),
        "wrap": (None, None),
        "name": ("x", "y"),
    }

    fig = Figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], transform=transform, coord_meta=coord_meta)
    fig.add_axes(ax)
    ax.grid(grid_type="contours")
    fig.savefig(tmp_path / "test.png")


@pytest.mark.usefixtures("tmp_plt_figure")
def test_plt_imshow_origin():
    # Regression test for a bug that caused origin to be set to upper when
    # plt.imshow was called.
    import matplotlib.pyplot as plt

    ax = plt.subplot(projection=WCS())
    plt.imshow(np.ones((2, 2)))
    assert ax.get_xlim() == (-0.5, 1.5)
    assert ax.get_ylim() == (-0.5, 1.5)


def test_ax_imshow_origin():
    # Regression test for a bug that caused origin to be set to upper when
    # ax.imshow was called with no origin
    fig = Figure()
    ax = fig.add_subplot(projection=WCS())
    ax.imshow(np.ones((2, 2)))
    assert ax.get_xlim() == (-0.5, 1.5)
    assert ax.get_ylim() == (-0.5, 1.5)


def test_grid_contour_large_spacing(tmp_path):
    # Regression test for a bug that caused a crash when grid was called and
    # didn't produce grid lines (due e.g. to too large spacing) and was then
    # called again.

    filename = tmp_path / "test.png"

    fig = Figure()
    ax = fig.add_subplot(projection=WCS())
    ax.set_xlim(-0.5, 1.5)
    ax.set_ylim(-0.5, 1.5)
    ax.coords[0].set_ticks(values=[] * u.one)

    ax.coords[0].grid(grid_type="contours")
    fig.savefig(filename)

    ax.coords[0].grid(grid_type="contours")
    fig.savefig(filename)


def test_contour_return():
    # Regression test for a bug that caused contour and contourf to return None
    # instead of the contour object.

    fig = Figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    fig.add_axes(ax)

    cset = ax.contour(np.arange(16).reshape(4, 4), transform=ax.get_transform("world"))
    assert isinstance(cset, QuadContourSet)

    cset = ax.contourf(np.arange(16).reshape(4, 4), transform=ax.get_transform("world"))
    assert isinstance(cset, QuadContourSet)


def test_contour_empty():
    # Regression test for a bug that caused contour to crash if no contours
    # were present.

    fig = Figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    fig.add_axes(ax)
    ax.contour(np.zeros((4, 4)), transform=ax.get_transform("world"))


def test_iterate_coords(ignore_matplotlibrc):
    # Regression test for a bug that caused ax.coords to return too few axes

    wcs3d = WCS(naxis=3)
    wcs3d.wcs.ctype = ["x", "y", "z"]
    wcs3d.wcs.cunit = ["deg", "deg", "km/s"]
    wcs3d.wcs.crpix = [614.5, 856.5, 333]
    wcs3d.wcs.cdelt = [6.25, 6.25, 23]
    wcs3d.wcs.crval = [0.0, 0.0, 1.0]

    fig = Figure()
    ax = fig.add_subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))

    x, y, z = ax.coords


def test_invalid_slices_errors(ignore_matplotlibrc):
    # Make sure that users get a clear message when specifying a WCS with
    # >2 dimensions without giving the 'slices' argument, or if the 'slices'
    # argument has too many/few elements.
    fig = Figure()
    wcs3d = WCS(naxis=3)
    wcs3d.wcs.ctype = ["x", "y", "z"]

    fig.add_subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))

    with pytest.raises(
        ValueError,
        match=r"WCS has more than 2 pixel dimensions, so 'slices' should be set",
    ):
        fig.add_subplot(1, 1, 1, projection=wcs3d)

    with pytest.raises(
        ValueError,
        match=(
            r"'slices' should have as many elements as WCS has pixel dimensions .should"
            r" be 3."
        ),
    ):
        fig.add_subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1, 2))

    wcs2d = WCS(naxis=2)
    wcs2d.wcs.ctype = ["x", "y"]

    fig.clear()
    ax = fig.add_subplot(1, 1, 1, projection=wcs2d)
    assert ax.frame_class is RectangularFrame
    fig.clear()
    ax = fig.add_subplot(1, 1, 1, projection=wcs2d, slices=("x", "y"))
    assert ax.frame_class is RectangularFrame
    fig.clear()
    ax = fig.add_subplot(1, 1, 1, projection=wcs2d, slices=("y", "x"))
    assert ax.frame_class is RectangularFrame
    fig.clear()
    ax = fig.add_subplot(1, 1, 1, projection=wcs2d, slices=["x", "y"])
    assert ax.frame_class is RectangularFrame
    fig.clear()
    ax = fig.add_subplot(1, 1, 1, projection=wcs2d, slices=(1, "x"))
    assert ax.frame_class is RectangularFrame1D

    wcs1d = WCS(naxis=1)
    wcs1d.wcs.ctype = ["x"]

    fig.clear()
    ax = fig.add_subplot(1, 1, 1, projection=wcs1d)
    assert ax.frame_class is RectangularFrame1D

    with pytest.raises(ValueError):
        fig.add_subplot(1, 1, 1, projection=wcs2d, slices=(1, "y"))


EXPECTED_REPR_1 = """
<CoordinatesMap with 3 world coordinates:

  index            aliases                type   ...    wrap   format_unit visible
  ----- ------------------------------ --------- ... --------- ----------- -------
      0                   distmod dist    scalar ...      None                  no
      1 pos.galactic.lon glon-car glon longitude ... 360.0 deg         deg     yes
      2 pos.galactic.lat glat-car glat  latitude ...      None         deg     yes

>
 """.strip()

EXPECTED_REPR_2 = """
<CoordinatesMap with 3 world coordinates:

  index            aliases                type   ...    wrap   format_unit visible
  ----- ------------------------------ --------- ... --------- ----------- -------
      0                   distmod dist    scalar ...      None                 yes
      1 pos.galactic.lon glon-car glon longitude ... 360.0 deg         deg     yes
      2 pos.galactic.lat glat-car glat  latitude ...      None         deg     yes

>
 """.strip()


def test_repr(ignore_matplotlibrc):
    # Unit test to make sure __repr__ looks as expected
    fig = Figure()
    wcs3d = WCS(GAL_HEADER)

    # Cube header has world coordinates as distance, lon, lat, so start off
    # by slicing in a way that we select just lon,lat:

    ax = fig.add_subplot(1, 1, 1, projection=wcs3d, slices=(1, "x", "y"))
    assert repr(ax.coords) == EXPECTED_REPR_1

    # Now slice in a way that all world coordinates are still present:

    fig.clear()
    ax = fig.add_subplot(1, 1, 1, projection=wcs3d, slices=("x", "y", 1))
    assert repr(ax.coords) == EXPECTED_REPR_2


@pytest.fixture
def time_spectral_wcs_2d():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ["FREQ", "TIME"]
    wcs.wcs.set()
    return wcs


def test_time_wcs(time_spectral_wcs_2d):
    # Regression test for a bug that caused WCSAxes to error when using a WCS
    # with a time axis.
    fig = Figure()
    fig.add_subplot(projection=time_spectral_wcs_2d)


@pytest.mark.skipif(TEX_UNAVAILABLE, reason="TeX is unavailable")
def test_simplify_labels_usetex(ignore_matplotlibrc, tmp_path):
    """Regression test for https://github.com/astropy/astropy/issues/8004."""
    mpl.rc("text", usetex=True)

    header = {
        "NAXIS": 2,
        "NAXIS1": 360,
        "NAXIS2": 180,
        "CRPIX1": 180.5,
        "CRPIX2": 90.5,
        "CRVAL1": 180.0,
        "CRVAL2": 0.0,
        "CDELT1": -2 * np.sqrt(2) / np.pi,
        "CDELT2": 2 * np.sqrt(2) / np.pi,
        "CTYPE1": "RA---MOL",
        "CTYPE2": "DEC--MOL",
        "RADESYS": "ICRS",
    }

    wcs = WCS(header)
    fig = Figure()
    ax = fig.add_subplot(frame_class=EllipticalFrame, projection=wcs)
    ax.set_xlim(-0.5, header["NAXIS1"] - 0.5)
    ax.set_ylim(-0.5, header["NAXIS2"] - 0.5)
    ax.coords[0].set_ticklabel(exclude_overlapping=True)
    ax.coords[1].set_ticklabel(exclude_overlapping=True)
    ax.coords[0].set_ticks(spacing=45 * u.deg)
    ax.coords[1].set_ticks(spacing=30 * u.deg)
    ax.grid()

    fig.savefig(tmp_path / "plot.png")


@pytest.mark.parametrize(
    "usetex, unicode_minus, label_str",
    [
        (True, True, "$-{}$"),
        (True, False, "$-{}$"),
        (False, True, "\N{MINUS SIGN}{}"),
        (False, False, "-{}"),
    ],
)
def test_simplify_labels_minus_sign(
    ignore_matplotlibrc, usetex, unicode_minus, label_str
):
    # Ensure minus signs aren't removed from the front of labels across a grid of configuration possibilities
    if usetex and TEX_UNAVAILABLE:
        pytest.skip("TeX is unavailable")

    ticklabels = TickLabels(frame=MagicMock())
    expected_labels = []
    for i in range(1, 6):
        label = label_str.format(i)
        ticklabels.add(
            axis="axis",
            world=0,
            angle=0,
            text=label,
            axis_displacement=0,
            data=(i, i),
        )
        expected_labels.append(label)

    with mpl.rc_context(
        rc={"text.usetex": usetex, "axes.unicode_minus": unicode_minus}
    ):
        ticklabels.simplify_labels()
    assert ticklabels.text["axis"] == expected_labels


@pytest.mark.parametrize("frame_class", [RectangularFrame, EllipticalFrame])
def test_set_labels_with_coords(ignore_matplotlibrc, frame_class):
    """Test if ``axis.set_xlabel()`` calls the correct ``coords[i]_set_axislabel()`` in a
    WCS plot. Regression test for https://github.com/astropy/astropy/issues/10435.
    """
    labels = ["RA", "Declination"]
    header = {
        "NAXIS": 2,
        "NAXIS1": 360,
        "NAXIS2": 180,
        "CRPIX1": 180.5,
        "CRPIX2": 90.5,
        "CRVAL1": 180.0,
        "CRVAL2": 0.0,
        "CDELT1": -2 * np.sqrt(2) / np.pi,
        "CDELT2": 2 * np.sqrt(2) / np.pi,
        "CTYPE1": "RA---AIT",
        "CTYPE2": "DEC--AIT",
    }

    wcs = WCS(header)
    fig = Figure()
    ax = fig.add_subplot(frame_class=frame_class, projection=wcs)
    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    assert ax.get_xlabel() == labels[0]
    assert ax.get_ylabel() == labels[1]
    for i in range(2):
        assert ax.coords[i].get_axislabel() == labels[i]


@pytest.mark.parametrize("atol", [0.2, 1.0e-8])
def test_bbox_size(atol):
    # Test for the size of a WCSAxes bbox (only have Matplotlib >= 3.0 now)
    extents = [11.38888888888889, 3.5, 576.0, 432.0]

    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8])
    canvas.figure.add_axes(ax)
    canvas.draw()
    renderer = fig.canvas.renderer
    ax_bbox = ax.get_tightbbox(renderer)

    # Enforce strict test only with reference Freetype version
    if atol < 0.1 and not FREETYPE_261:
        pytest.xfail(
            "Exact BoundingBox dimensions are only ensured with FreeType 2.6.1"
        )
    assert np.allclose(ax_bbox.extents, extents, atol=atol)


def test_wcs_type_transform_regression():
    fig = Figure()
    wcs = WCS(TARGET_HEADER)
    sliced_wcs = SlicedLowLevelWCS(wcs, np.s_[1:-1, 1:-1])
    ax = fig.add_subplot(1, 1, 1, projection=wcs)
    ax.get_transform(sliced_wcs)

    high_wcs = HighLevelWCSWrapper(sliced_wcs)
    ax.get_transform(sliced_wcs)


def test_multiple_draws_grid_contours(tmp_path):
    fig = Figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS())
    ax.grid(color="black", grid_type="contours")
    fig.savefig(tmp_path / "plot.png")
    fig.savefig(tmp_path / "plot.png")


def test_get_coord_range_nan_regression():
    # Test to make sure there is no internal casting of NaN to integers
    # NumPy 1.24 raises a RuntimeWarning if a NaN is cast to an integer

    fig = Figure()
    wcs = WCS(TARGET_HEADER)
    wcs.wcs.crval[0] = 0  # Re-position the longitude wrap to the middle
    ax = fig.add_subplot(1, 1, 1, projection=wcs)

    # Set the Y limits within valid latitudes/declinations
    ax.set_ylim(300, 500)

    # Set the X limits within valid longitudes/RAs, so the world coordinates have no NaNs
    ax.set_xlim(300, 700)
    assert np.allclose(
        ax.coords.get_coord_range(),
        np.array(
            [
                (-123.5219272110385, 122.49684897692201),
                (-44.02289164685554, 44.80732766607591),
            ]
        ),
    )

    # Extend the X limits to include invalid longitudes/RAs, so the world coordinates have NaNs
    ax.set_xlim(0, 700)
    assert np.allclose(
        ax.coords.get_coord_range(),
        np.array(
            [(-131.3193386797236, 180.0), (-44.02289164685554, 44.80732766607591)]
        ),
    )


def test_imshow_error():
    fig = Figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS())
    with pytest.raises(ValueError, match="Cannot use images with origin='upper"):
        ax.imshow(np.ones(100).reshape(10, 10), origin="upper")


def test_label_setting():
    fig = Figure()
    ax = fig.add_subplot(1, 1, 1, projection=WCS())
    # Check both xlabel and label kwargs work
    ax.set_xlabel(xlabel="label")
    ax.set_xlabel(label="label")
    # Check no label errors:
    with pytest.raises(
        TypeError, match=r"set_xlabel\(\) missing 1 required positional argument"
    ):
        ax.set_xlabel()

    # Check both xlabel and label kwargs work
    ax.set_ylabel(ylabel="label")
    ax.set_ylabel(label="label")
    # Check no label errors:
    with pytest.raises(
        TypeError, match=r"set_ylabel\(\) missing 1 required positional argument"
    ):
        ax.set_ylabel()


def test_invisible_bbox():
    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = fig.add_subplot(1, 1, 1, projection=WCS())

    assert ax.get_tightbbox(fig.canvas.get_renderer()) is not None
    ax.set_visible(False)
    assert ax.get_tightbbox(fig.canvas.get_renderer()) is None


def test_get_axislabel_default():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "RA---TAN", "DEC--TAN"

    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = fig.add_subplot(1, 1, 1, projection=wcs)
    assert ax.coords[0].get_axislabel() == "pos.eq.ra"
    assert ax.coords[1].get_axislabel() == "pos.eq.dec"

    # Make sure that setting axis labels explicitly works, including to an
    # empty string
    ax.coords[0].set_axislabel("Right Ascension")
    ax.coords[1].set_axislabel("")
    assert ax.coords[0].get_axislabel() == "Right Ascension"
    assert ax.coords[1].get_axislabel() == ""


def test_plot_coord_slicing(ignore_matplotlibrc):
    # Regression test to ensure plot_coord does not raise a NotImplementedError
    # after slicing coordinates (issue #10254).

    cube_header = get_pkg_data_filename("data/cube_header")
    cube_header = fits.Header.fromtextfile(cube_header)

    fig = Figure(figsize=(6, 6))
    ax = fig.add_subplot(projection=WCS(cube_header), slices=("x", "y", 0))

    c = SkyCoord(52 * u.deg, 30.5 * u.deg)
    ax.plot_coord(c, "o")


SIMPLIFY_CASES = [
    (["13d14m15s", "13d14m15s"], ["13d14m15s", "15s"]),
    (["13d14m5s", "13d14m15s"], ["13d14m5s", "15s"]),
    (
        ["−29°25'55\"", "−29°25'54\"", "−29°25'53\"", "−29°25'52\"", "−29°25'51\""],
        ["−29°25'55\"", '54"', '53"', '52"', '51"'],
    ),
    (
        [
            r"0$\mathregular{^h}$",
            r"10$\mathregular{^h}$",
            r"10$\mathregular{^h}$",
            r"5$\mathregular{^h}$",
        ],
        [
            r"0$\mathregular{^h}$",
            r"10$\mathregular{^h}$",
            r"10$\mathregular{^h}$",
            r"5$\mathregular{^h}$",
        ],
    ),
    (
        [
            r"0$\mathregular{^h}$10$\mathregular{^m}$5$\mathregular{^m}$",
            r"0$\mathregular{^h}$10$\mathregular{^m}$10$\mathregular{^m}$",
            r"0$\mathregular{^h}$10$\mathregular{^m}$15$\mathregular{^m}$",
            r"0$\mathregular{^h}$10$\mathregular{^m}$15$\mathregular{^m}$",
        ],
        [
            r"0$\mathregular{^h}$10$\mathregular{^m}$5$\mathregular{^m}$",
            r"10$\mathregular{^m}$",
            r"15$\mathregular{^m}$",
            r"15$\mathregular{^m}$",
        ],
    ),
    (
        [
            "$17^{\\mathrm{h}}47^{\\mathrm{m}}53.8^{\\mathrm{s}}$",
            "$17^{\\mathrm{h}}47^{\\mathrm{m}}53.6^{\\mathrm{s}}$",
        ],
        [
            "$17^{\\mathrm{h}}47^{\\mathrm{m}}53.8^{\\mathrm{s}}$",
            "$53.6^{\\mathrm{s}}$",
        ],
    ),
]


@pytest.mark.parametrize(("before", "after"), SIMPLIFY_CASES)
def test_simplify_cases(before, after):
    # Regression test for a bug that caused the simplification to not work
    # correctly in the presence of dollar signs in LaTeX strings.

    ticklabels = TickLabels(frame=MagicMock())
    expected_labels = []

    for i, label in enumerate(before):
        ticklabels.add(
            axis="axis",
            world=0,
            angle=0,
            text=label,
            axis_displacement=i,
            data=(i, i),
        )

    ticklabels.simplify_labels()
    assert ticklabels.text["axis"] == after


def test_get_transform_unit_mismatch():
    """
    Regression test for a bug that caused get_transform to ignore differences
    in WCS units.

    https://github.com/astropy/astropy/issues/18246
    """

    wcs_deg = WCS(naxis=2, preserve_units=True)
    wcs_deg.wcs.ctype = "RA---TAN", "DEC--TAN"
    wcs_deg.wcs.crval = 20, 30
    wcs_deg.wcs.cunit = "deg", "deg"
    wcs_deg.wcs.crpix = 1, 1
    wcs_deg.wcs.cdelt = 1 / 60, 1 / 60

    wcs_arcmin = WCS(naxis=2, preserve_units=True)
    wcs_arcmin.wcs.ctype = "RA---TAN", "DEC--TAN"
    wcs_arcmin.wcs.crval = 1200, 1800
    wcs_arcmin.wcs.cunit = "arcmin", "arcmin"
    wcs_arcmin.wcs.crpix = 1, 1
    wcs_arcmin.wcs.cdelt = 1, 1

    fig = Figure(figsize=(6, 6))
    ax = fig.add_subplot(projection=wcs_arcmin)
    transform1 = ax.get_transform(wcs_arcmin)
    transform2 = ax.get_transform(wcs_deg)

    # Since the two WCS are equivalent, the returned transforms should also
    # be equivalent

    rng = np.random.default_rng(12345)
    pixels = rng.uniform(0, 100, (2, 100))

    assert_allclose(transform1.transform(pixels), transform2.transform(pixels))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Utilities for generating documentation for unit definition modules.

None of the functions in the module are meant for use outside of the
package.
"""

import keyword
import re
import unicodedata
from collections.abc import Iterable, Mapping
from io import StringIO

from .core import NamedUnit, PrefixUnit, Unit, UnitBase


def _get_first_sentence(s: str) -> str:
    """
    Get the first sentence from a string and remove any carriage
    returns.
    """
    x = re.match(r".*?\S\.\s", s)
    if x is not None:
        s = x.group(0)
    return s.replace("\n", " ")


def _summarize_units(
    units: Iterable[NamedUnit],
    have_prefixes: set[str],
    docstring: StringIO,
    template: str,
) -> str:
    for unit in sorted(units, key=lambda x: x.name.lower()):
        represents = ""
        if isinstance(unit, Unit):
            represents = f":math:`{unit._represents.to_string('latex')[1:-1]}`"
        docstring.write(
            template.format(
                unit,
                _get_first_sentence(unit.__doc__).strip(),
                represents,
                ", ".join(f"``{x}``" for x in unit.aliases),
                "Yes" if unit.name in have_prefixes else "No",
            )
        )
    return docstring.getvalue()


def generate_unit_summary(namespace: Mapping[str, object]) -> str:
    """
    Generates a summary of units from a given namespace.  This is used
    to generate the docstring for the modules that define the actual
    units.

    Parameters
    ----------
    namespace : dict
        A namespace containing units.

    Returns
    -------
    docstring : str
        A docstring containing a summary table of the units.
    """
    units = []
    have_prefixes = set()
    for key, val in namespace.items():
        if not isinstance(val, UnitBase):
            continue
        if not isinstance(val, NamedUnit):
            raise TypeError(f"{key!r} must be defined with 'def_unit()'")
        if key == val.name:  # Skip aliases
            if isinstance(val, PrefixUnit):
                # This will return the root unit that is scaled by the prefix
                # attached to it
                have_prefixes.add(val._represents.bases[0].name)
            else:
                units.append(val)
    docstring = StringIO()
    docstring.write(
        """
.. list-table:: Available Units
   :header-rows: 1
   :widths: 10 20 20 20 1

   * - Unit
     - Description
     - Represents
     - Aliases
     - SI Prefixes
"""
    )
    template = """
   * - ``{}``
     - {}
     - {}
     - {}
     - {}
"""
    return _summarize_units(units, have_prefixes, docstring, template)


def generate_prefixonly_unit_summary(namespace: Mapping[str, object]) -> str:
    """
    Generates table entries for units in a namespace that are just prefixes
    without the base unit.  Note that this is intended to be used *after*
    `generate_unit_summary` and therefore does not include the table header.

    Parameters
    ----------
    namespace : dict
        A namespace containing units that are prefixes but do *not* have the
        base unit in their namespace.

    Returns
    -------
    docstring : str
        A docstring containing a summary table of the units.
    """
    non_prefixed_units = {
        unit.represents.bases[0]
        for unit in namespace.values()
        if isinstance(unit, PrefixUnit)
    }
    template = """
   * - Prefixes for ``{}``
     - {} prefixes
     - {}
     - {}
     - Only
"""
    return _summarize_units(non_prefixed_units, set(), StringIO(), template)


def generate_dunder_all(namespace: Mapping[str, object]) -> list[str]:
    return [
        name
        for name, value in namespace.items()
        if (
            isinstance(value, UnitBase)
            and name.isidentifier()
            and not keyword.iskeyword(name)
            and name == unicodedata.normalize("NFKC", name)
        )
    ]



# Licensed under a 3-clause BSD style license - see LICENSE.rst


# TODO: this file should be refactored to use a more thread-safe and
# race-condition-safe lockfile mechanism.

import datetime
import os
import stat
import warnings
import xmlrpc.client as xmlrpc
from contextlib import suppress
from pathlib import Path
from urllib.parse import urlparse

from astropy import log
from astropy.utils.data import get_readable_fileobj

from .errors import SAMPHubError, SAMPWarning


def read_lockfile(lockfilename):
    """
    Read in the lockfile given by ``lockfilename`` into a dictionary.
    """
    # lockfilename may be a local file or a remote URL, but
    # get_readable_fileobj takes care of this.
    lockfiledict = {}
    with get_readable_fileobj(lockfilename) as f:
        for line in f:
            if not line.startswith("#"):
                kw, val = line.split("=")
                lockfiledict[kw.strip()] = val.strip()
    return lockfiledict


def write_lockfile(lockfilename, lockfiledict):
    lockfile = open(lockfilename, "w")
    lockfile.close()
    os.chmod(lockfilename, stat.S_IREAD + stat.S_IWRITE)

    lockfile = open(lockfilename, "w")
    now_iso = datetime.datetime.now().isoformat()
    lockfile.write(f"# SAMP lockfile written on {now_iso}\n")
    lockfile.write("# Standard Profile required keys\n")
    lockfile.writelines(f"{key}={value}\n" for key, value in lockfiledict.items())
    lockfile.close()


def create_lock_file(lockfilename=None, mode=None, hub_id=None, hub_params=None):
    # Remove lock-files of dead hubs
    remove_garbage_lock_files()

    lockfiledir = ""

    # CHECK FOR SAMP_HUB ENVIRONMENT VARIABLE
    if "SAMP_HUB" in os.environ:
        # For the time being I assume just the std profile supported.
        if os.environ["SAMP_HUB"].startswith("std-lockurl:"):
            lockfilename = os.environ["SAMP_HUB"][len("std-lockurl:") :]
            lockfile_parsed = urlparse(lockfilename)

            if lockfile_parsed[0] != "file":
                warnings.warn(
                    f"Unable to start a Hub with lockfile {lockfilename}. "
                    "Start-up process aborted.",
                    SAMPWarning,
                )
                return False
            else:
                lockfilename = lockfile_parsed[2]
    else:
        # If it is a fresh Hub instance
        if lockfilename is None:
            log.debug("Running mode: " + mode)

            if mode == "single":
                lockfilename = os.path.join(Path.home(), ".samp")
            else:
                lockfiledir = os.path.join(Path.home(), ".samp-1")

                # If missing create .samp-1 directory
                try:
                    os.mkdir(lockfiledir)
                except OSError:
                    pass  # directory already exists
                finally:
                    os.chmod(lockfiledir, stat.S_IREAD + stat.S_IWRITE + stat.S_IEXEC)

                lockfilename = os.path.join(lockfiledir, f"samp-hub-{hub_id}")

        else:
            log.debug("Running mode: multiple")

    hub_is_running, lockfiledict = check_running_hub(lockfilename)

    if hub_is_running:
        warnings.warn(
            "Another SAMP Hub is already running. Start-up process aborted.",
            SAMPWarning,
        )
        return False

    log.debug("Lock-file: " + lockfilename)

    write_lockfile(lockfilename, hub_params)

    return lockfilename


def get_main_running_hub():
    """
    Get either the hub given by the environment variable SAMP_HUB, or the one
    given by the lockfile .samp in the user home directory.
    """
    hubs = get_running_hubs()

    if not hubs:
        raise SAMPHubError("Unable to find a running SAMP Hub.")

    # CHECK FOR SAMP_HUB ENVIRONMENT VARIABLE
    if "SAMP_HUB" in os.environ:
        # For the time being I assume just the std profile supported.
        if os.environ["SAMP_HUB"].startswith("std-lockurl:"):
            lockfilename = os.environ["SAMP_HUB"][len("std-lockurl:") :]
        else:
            raise SAMPHubError("SAMP Hub profile not supported.")
    else:
        lockfilename = os.path.join(Path.home(), ".samp")

    return hubs[lockfilename]


def get_running_hubs():
    """
    Return a dictionary containing the lock-file contents of all the currently
    running hubs (single and/or multiple mode).

    The dictionary format is:

    ``{<lock-file>: {<token-name>: <token-string>, ...}, ...}``

    where ``{<lock-file>}`` is the lock-file name, ``{<token-name>}`` and
    ``{<token-string>}`` are the lock-file tokens (name and content).

    Returns
    -------
    running_hubs : dict
        Lock-file contents of all the currently running hubs.
    """
    hubs = {}
    lockfilename = ""

    # HUB SINGLE INSTANCE MODE

    # CHECK FOR SAMP_HUB ENVIRONMENT VARIABLE
    if "SAMP_HUB" in os.environ:
        # For the time being I assume just the std profile supported.
        if os.environ["SAMP_HUB"].startswith("std-lockurl:"):
            lockfilename = os.environ["SAMP_HUB"][len("std-lockurl:") :]
    else:
        lockfilename = os.path.join(Path.home(), ".samp")

    hub_is_running, lockfiledict = check_running_hub(lockfilename)

    if hub_is_running:
        hubs[lockfilename] = lockfiledict

    # HUB MULTIPLE INSTANCE MODE

    lockfiledir = ""

    lockfiledir = os.path.join(Path.home(), ".samp-1")

    if os.path.isdir(lockfiledir):
        for filename in os.listdir(lockfiledir):
            if filename.startswith("samp-hub"):
                lockfilename = os.path.join(lockfiledir, filename)
                hub_is_running, lockfiledict = check_running_hub(lockfilename)
                if hub_is_running:
                    hubs[lockfilename] = lockfiledict

    return hubs


def check_running_hub(lockfilename):
    """
    Test whether a hub identified by ``lockfilename`` is running or not.

    Parameters
    ----------
    lockfilename : str
        Lock-file name (path + file name) of the Hub to be tested.

    Returns
    -------
    is_running : bool
        Whether the hub is running
    hub_params : dict
        If the hub is running this contains the parameters from the lockfile
    """
    is_running = False
    lockfiledict = {}

    # Check whether a lockfile already exists
    try:
        lockfiledict = read_lockfile(lockfilename)
    except OSError:
        return is_running, lockfiledict

    if "samp.hub.xmlrpc.url" in lockfiledict:
        try:
            proxy = xmlrpc.ServerProxy(
                lockfiledict["samp.hub.xmlrpc.url"].replace("\\", ""), allow_none=1
            )
            proxy.samp.hub.ping()
            is_running = True
        except xmlrpc.ProtocolError:
            # There is a protocol error (e.g. for authentication required),
            # but the server is alive
            is_running = True
        except OSError:
            pass

    return is_running, lockfiledict


def remove_garbage_lock_files():
    lockfilename = ""

    # HUB SINGLE INSTANCE MODE

    lockfilename = os.path.join(Path.home(), ".samp")

    hub_is_running, lockfiledict = check_running_hub(lockfilename)

    if not hub_is_running:
        # If lockfilename belongs to a dead hub, then it is deleted
        if os.path.isfile(lockfilename):
            with suppress(OSError):
                os.remove(lockfilename)

    # HUB MULTIPLE INSTANCE MODE

    lockfiledir = os.path.join(Path.home(), ".samp-1")

    if os.path.isdir(lockfiledir):
        for filename in os.listdir(lockfiledir):
            if filename.startswith("samp-hub"):
                lockfilename = os.path.join(lockfiledir, filename)
                hub_is_running, lockfiledict = check_running_hub(lockfilename)
                if not hub_is_running:
                    # If lockfilename belongs to a dead hub, then it is deleted
                    if os.path.isfile(lockfilename):
                        with suppress(OSError):
                            os.remove(lockfilename)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from astropy import units as u
from astropy.coordinates import representation as r
from astropy.coordinates.attributes import (
    EarthLocationAttribute,
    QuantityAttribute,
    TimeAttribute,
)
from astropy.coordinates.baseframe import (
    BaseCoordinateFrame,
    RepresentationMapping,
    base_doc,
)
from astropy.utils.decorators import format_doc

__all__ = ["AltAz"]


_90DEG = 90 * u.deg

doc_components = """
    az : `~astropy.coordinates.Angle`, optional, keyword-only
        The Azimuth for this object (``alt`` must also be given and
        ``representation`` must be None).
    alt : `~astropy.coordinates.Angle`, optional, keyword-only
        The Altitude for this object (``az`` must also be given and
        ``representation`` must be None).
    distance : `~astropy.units.Quantity` ['length'], optional, keyword-only
        The Distance for this object along the line-of-sight.

    pm_az_cosalt : `~astropy.units.Quantity` ['angular speed'], optional, keyword-only
        The proper motion in azimuth (including the ``cos(alt)`` factor) for
        this object (``pm_alt`` must also be given).
    pm_alt : `~astropy.units.Quantity` ['angular speed'], optional, keyword-only
        The proper motion in altitude for this object (``pm_az_cosalt`` must
        also be given).
    radial_velocity : `~astropy.units.Quantity` ['speed'], optional, keyword-only
        The radial velocity of this object."""

doc_footer = """
    Other parameters
    ----------------
    obstime : `~astropy.time.Time`
        The time at which the observation is taken.  Used for determining the
        position and orientation of the Earth.
    location : `~astropy.coordinates.EarthLocation`
        The location on the Earth.  This can be specified either as an
        `~astropy.coordinates.EarthLocation` object or as anything that can be
        transformed to an `~astropy.coordinates.ITRS` frame.
    pressure : `~astropy.units.Quantity` ['pressure']
        The atmospheric pressure as an `~astropy.units.Quantity` with pressure
        units.  This is necessary for performing refraction corrections.
        Setting this to 0 (the default) will disable refraction calculations
        when transforming to/from this frame.
    temperature : `~astropy.units.Quantity` ['temperature']
        The ground-level temperature as an `~astropy.units.Quantity` in
        deg C.  This is necessary for performing refraction corrections.
    relative_humidity : `~astropy.units.Quantity` ['dimensionless'] or number
        The relative humidity as a dimensionless quantity between 0 to 1.
        This is necessary for performing refraction corrections.
    obswl : `~astropy.units.Quantity` ['length']
        The average wavelength of observations as an `~astropy.units.Quantity`
         with length units.  This is necessary for performing refraction
         corrections.

    Notes
    -----
    The refraction model is based on that implemented in ERFA, which is fast
    but becomes inaccurate for altitudes below about 5 degrees.  Near and below
    altitudes of 0, it can even give meaningless answers, and in this case
    transforming to AltAz and back to another frame can give highly discrepant
    results.  For much better numerical stability, leave the ``pressure`` at
    ``0`` (the default), thereby disabling the refraction correction and
    yielding "topocentric" horizontal coordinates.
    """


@format_doc(base_doc, components=doc_components, footer=doc_footer)
class AltAz(BaseCoordinateFrame):
    """
    A coordinate or frame in the Altitude-Azimuth system (Horizontal
    coordinates) with respect to the WGS84 ellipsoid.  Azimuth is oriented
    East of North (i.e., N=0, E=90 degrees).  Altitude is also known as
    elevation angle, so this frame is also in the Azimuth-Elevation system.

    This frame is assumed to *include* refraction effects if the ``pressure``
    frame attribute is non-zero.

    The frame attributes are listed under **Other Parameters**, which are
    necessary for transforming from AltAz to some other system.
    """

    frame_specific_representation_info = {
        r.SphericalRepresentation: [
            RepresentationMapping("lon", "az"),
            RepresentationMapping("lat", "alt"),
        ]
    }

    default_representation = r.SphericalRepresentation
    default_differential = r.SphericalCosLatDifferential

    obstime = TimeAttribute(
        default=None, doc="The reference time (e.g., time of observation)"
    )
    location = EarthLocationAttribute(
        default=None, doc="The location on Earth of the observer"
    )
    pressure = QuantityAttribute(default=0, unit=u.hPa, doc="The atmospheric pressure")
    temperature = QuantityAttribute(
        default=0, unit=u.deg_C, doc="The ground-level temperature"
    )
    relative_humidity = QuantityAttribute(
        default=0, unit=u.dimensionless_unscaled, doc="The relative humidity"
    )
    obswl = QuantityAttribute(
        default=1 * u.micron,
        unit=u.micron,
        doc="The average wavelength of observations",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def secz(self):
        """
        Secant of the zenith angle for this coordinate, a common estimate of
        the airmass.
        """
        return 1 / np.sin(self.alt)

    @property
    def zen(self):
        """
        The zenith angle (or zenith distance / co-altitude) for this coordinate.
        """
        return _90DEG.to(self.alt.unit) - self.alt


# self-transform defined in icrs_observed_transforms.py



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Testing :mod:`astropy.cosmology.FLRW`."""

from typing import final

import pytest

from astropy.cosmology import FLRW
from astropy.cosmology._src.core import _COSMOLOGY_CLASSES, dataclass_decorator
from astropy.cosmology._src.tests.helper import get_redshift_methods
from astropy.cosmology._src.tests.test_core import invalid_zs
from astropy.utils.compat.optional_deps import HAS_SCIPY

from .test_base import FLRWTest


@dataclass_decorator
class SubFLRW(FLRW):
    def w(self, z):
        return super().w(z)


@final
class TestFLRW(FLRWTest):
    """Test :class:`astropy.cosmology.FLRW`."""

    abstract_w = True

    def setup_class(self):
        """
        Setup for testing.
        FLRW is abstract, so tests are done on a subclass.
        """
        super().setup_class(self)

        # make sure SubCosmology is known
        _COSMOLOGY_CLASSES["SubFLRW"] = SubFLRW

        self.cls = SubFLRW

    def teardown_class(self):
        super().teardown_class(self)
        _COSMOLOGY_CLASSES.pop("SubFLRW", None)

    # ===============================================================
    # Method & Attribute Tests

    # ---------------------------------------------------------------
    # Methods

    def test_w(self, cosmo):
        """Test abstract :meth:`astropy.cosmology.FLRW.w`."""
        with pytest.raises(NotImplementedError, match="not implemented"):
            cosmo.w(1)

    def test_Otot(self, cosmo):
        """Test :meth:`astropy.cosmology.FLRW.Otot`."""
        exception = NotImplementedError if HAS_SCIPY else ModuleNotFoundError
        with pytest.raises(exception):
            assert cosmo.Otot(1)

    def test_efunc_vs_invefunc(self, cosmo):
        """
        Test that efunc and inv_efunc give inverse values.
        Here they just fail b/c no ``w(z)`` or no scipy.
        """
        exception = NotImplementedError if HAS_SCIPY else ModuleNotFoundError

        with pytest.raises(exception):
            cosmo.efunc(0.5)

        with pytest.raises(exception):
            cosmo.inv_efunc(0.5)

    @pytest.mark.skip(reason="w(z) is abstract")
    def test_luminosity_distance_pandas(self, cosmo):
        """Test :meth:`astropy.cosmology.FLRW.luminosity_distance`."""

    _FLRW_redshift_methods = get_redshift_methods(
        FLRW, include_private=True, include_z2=False
    ) - {"w"}

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy is not installed")
    @pytest.mark.parametrize("z, exc", invalid_zs)
    @pytest.mark.parametrize("method", sorted(_FLRW_redshift_methods))
    def test_redshift_method_bad_input(self, cosmo, method, z, exc):
        """Test all the redshift methods for bad input."""
        with pytest.raises(exc):
            getattr(cosmo, method)(z)

    # ===============================================================
    # Usage Tests

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy required for this test.")
    @pytest.mark.parametrize("method", ("Om", "Ode", "w", "de_density_scale"))
    def test_distance_broadcast(self, cosmo, method):
        with pytest.raises(NotImplementedError):
            super().test_distance_broadcast(cosmo, method)

    @pytest.mark.skip(reason="w(z) is abstract")
    def test_comoving_distance_1arg_equal_to_2arg(self, cosmo):
        """Test :meth:`astropy.cosmology.FLRW.luminosity_distance`."""

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy required for this test.")
    @pytest.mark.parametrize(
        ("args", "kwargs", "expected"),
        [((70, 0.27, 0.73), {"Tcmb0": 3.0, "Ob0": 0.03}, None)],
    )
    def test_comoving_distance_example(self, cosmo_cls, args, kwargs, expected):
        with pytest.raises(NotImplementedError):
            super().test_comoving_distance_example(cosmo_cls, args, kwargs, expected)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


# This file defines the AngleFormatterLocator class which is a class that
# provides both a method for a formatter and one for a locator, for a given
# label spacing. The advantage of keeping the two connected is that we need to
# make sure that the formatter can correctly represent the spacing requested and
# vice versa. For example, a format of dd:mm cannot work with a tick spacing
# that is not a multiple of one arcminute.

import re
import warnings

import numpy as np
from matplotlib import rcParams
from matplotlib.ticker import Formatter

from astropy import units as u
from astropy.coordinates import Angle
from astropy.units import UnitsError

DMS_RE = re.compile("^dd(:mm(:ss(.(s)+)?)?)?$")
HMS_RE = re.compile("^hh(:mm(:ss(.(s)+)?)?)?$")
DDEC_RE = re.compile("^d(.(d)+)?$")
DMIN_RE = re.compile("^m(.(m)+)?$")
DSEC_RE = re.compile("^s(.(s)+)?$")
SCAL_RE = re.compile("^x(.(x)+)?$")


# Units with custom representations - see the note where it is used inside
# AngleFormatterLocator.formatter for more details.

CUSTOM_UNITS = {
    u.degree: u.def_unit(
        "custom_degree",
        represents=u.degree,
        format={"generic": "\xb0", "latex": r"^\circ", "unicode": "°"},
    ),
    u.arcmin: u.def_unit(
        "custom_arcmin",
        represents=u.arcmin,
        format={"generic": "'", "latex": r"^\prime", "unicode": "′"},
    ),
    u.arcsec: u.def_unit(
        "custom_arcsec",
        represents=u.arcsec,
        format={"generic": '"', "latex": r"^{\prime\prime}", "unicode": "″"},
    ),
    u.hourangle: u.def_unit(
        "custom_hourangle",
        represents=u.hourangle,
        format={
            "generic": "h",
            "latex": r"^{\mathrm{h}}",
            "unicode": r"$\mathregular{^h}$",
        },
    ),
}


def _fix_minus(labels: list[str], /) -> list[str]:
    # correctly support axes.unicode_minus, but do it in a
    # way that preserves arbitrary separators: only fix the leading character
    # see https://github.com/astropy/astropy/issues/15898
    return [Formatter.fix_minus(s[0]) + s[1:] for s in labels]


class BaseFormatterLocator:
    """
    A joint formatter/locator.
    """

    def __init__(
        self,
        values=None,
        number=None,
        spacing=None,
        format=None,
        unit=None,
        format_unit=None,
    ):
        if len([x for x in (values, number, spacing) if x is None]) < 2:
            raise ValueError("At most one of values/number/spacing can be specified")

        self._unit = unit
        self._format_unit = format_unit or unit

        if values is not None:
            self.values = values
        elif number is not None:
            self.number = number
        elif spacing is not None:
            self.spacing = spacing
        else:
            self.number = 5

        self.format = format

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, values):
        if not isinstance(values, u.Quantity) or (not values.ndim == 1):
            raise TypeError("values should be an astropy.units.Quantity array")
        if not values.unit.is_equivalent(self._unit):
            raise UnitsError(
                "value should be in units compatible with "
                f"coordinate units ({self._unit}) but found {values.unit}"
            )
        self._number = None
        self._spacing = None
        self._values = values

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, number):
        self._number = number
        self._spacing = None
        self._values = None

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, spacing):
        self._number = None
        self._spacing = spacing
        self._values = None

    def minor_locator(self, spacing, frequency, value_min, value_max):
        if self.values is not None:
            return [] * self._unit

        minor_spacing = spacing.value / frequency
        values = self._locate_values(value_min, value_max, minor_spacing)
        index = np.where((values % frequency) == 0)
        index = index[0][0]
        values = np.delete(values, np.s_[index::frequency])
        return values * minor_spacing * self._unit

    @property
    def format_unit(self):
        return self._format_unit

    @format_unit.setter
    def format_unit(self, unit):
        self._format_unit = u.Unit(unit)

    @staticmethod
    def _locate_values(value_min, value_max, spacing):
        imin = np.ceil(value_min / spacing)
        imax = np.floor(value_max / spacing)
        return np.arange(imin, imax + 1, dtype=int)


class AngleFormatterLocator(BaseFormatterLocator):
    """
    A joint formatter/locator.

    Parameters
    ----------
    number : int, optional
        Number of ticks.
    """

    def __init__(
        self,
        values=None,
        number=None,
        spacing=None,
        format=None,
        unit=None,
        decimal=None,
        format_unit=None,
        show_decimal_unit=True,
    ):
        if unit is None:
            unit = u.degree

        if format_unit is None:
            format_unit = unit

        if format_unit not in (u.degree, u.hourangle, u.hour):
            if decimal is False:
                raise UnitsError(
                    "Units should be degrees or hours when using non-decimal"
                    " (sexagesimal) mode"
                )

        self._decimal = decimal
        self._sep = None
        self.show_decimal_unit = show_decimal_unit
        self._alwayssign = False

        super().__init__(
            values=values,
            number=number,
            spacing=spacing,
            format=format,
            unit=unit,
            format_unit=format_unit,
        )

    @property
    def decimal(self):
        decimal = self._decimal
        if self.format_unit not in (u.degree, u.hourangle, u.hour):
            if self._decimal is None:
                decimal = True
            elif self._decimal is False:
                raise UnitsError(
                    "Units should be degrees or hours when using non-decimal"
                    " (sexagesimal) mode"
                )
        elif self._decimal is None:
            decimal = False
        return decimal

    @decimal.setter
    def decimal(self, value):
        self._decimal = value

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, spacing):
        if spacing is not None and (
            not isinstance(spacing, u.Quantity) or spacing.unit.physical_type != "angle"
        ):
            raise TypeError(
                "spacing should be an astropy.units.Quantity "
                "instance with units of angle"
            )
        self._number = None
        self._spacing = spacing
        self._values = None

    @property
    def sep(self):
        return self._sep

    @sep.setter
    def sep(self, separator):
        self._sep = separator

    @property
    def format(self):
        return self._format

    @format.setter
    def format(self, value):
        self._format = value

        if value is None:
            return

        self._alwayssign = value.startswith("+")
        if self._alwayssign:
            value = value[1:]

        if DMS_RE.match(value) is not None:
            self._decimal = False
            self._format_unit = u.degree
            if "." in value:
                self._precision = len(value) - value.index(".") - 1
                self._fields = 3
            else:
                self._precision = 0
                self._fields = value.count(":") + 1
        elif HMS_RE.match(value) is not None:
            self._decimal = False
            self._format_unit = u.hourangle
            if "." in value:
                self._precision = len(value) - value.index(".") - 1
                self._fields = 3
            else:
                self._precision = 0
                self._fields = value.count(":") + 1
        elif DDEC_RE.match(value) is not None:
            self._decimal = True
            self._format_unit = u.degree
            self._fields = 1
            if "." in value:
                self._precision = len(value) - value.index(".") - 1
            else:
                self._precision = 0
        elif DMIN_RE.match(value) is not None:
            self._decimal = True
            self._format_unit = u.arcmin
            self._fields = 1
            if "." in value:
                self._precision = len(value) - value.index(".") - 1
            else:
                self._precision = 0
        elif DSEC_RE.match(value) is not None:
            self._decimal = True
            self._format_unit = u.arcsec
            self._fields = 1
            if "." in value:
                self._precision = len(value) - value.index(".") - 1
            else:
                self._precision = 0
        else:
            raise ValueError(f"Invalid format: {value}")

        if self.spacing is not None and self.spacing < self.base_spacing:
            warnings.warn("Spacing is too small - resetting spacing to match format")
            self.spacing = self.base_spacing

        if self.spacing is not None:
            ratio = (self.spacing / self.base_spacing).decompose().value
            remainder = ratio - np.round(ratio)

            if abs(remainder) > 1.0e-10:
                warnings.warn(
                    "Spacing is not a multiple of base spacing - resetting spacing to"
                    " match format"
                )
                self.spacing = self.base_spacing * max(1, round(ratio))

    @property
    def base_spacing(self):
        if self.decimal:
            spacing = self._format_unit / (10.0**self._precision)

        else:
            if self._fields == 1:
                spacing = 1.0 * u.degree
            elif self._fields == 2:
                spacing = 1.0 * u.arcmin
            elif self._fields == 3:
                if self._precision == 0:
                    spacing = 1.0 * u.arcsec
                else:
                    spacing = u.arcsec / (10.0**self._precision)

        if self._format_unit is u.hourangle:
            spacing *= 15

        return spacing

    def locator(self, value_min, value_max):
        if self.values is not None:
            # values were manually specified
            return self.values, 1.1 * u.arcsec

        else:
            # In the special case where value_min is the same as value_max, we
            # don't locate any ticks. This can occur for example when taking a
            # slice for a cube (along the dimension sliced). We return a
            # non-zero spacing in case the caller needs to format a single
            # coordinate, e.g. for mousover.
            if value_min == value_max:
                return [] * self._unit, 1 * u.arcsec

            if self.spacing is not None:
                # spacing was manually specified
                spacing_value = self.spacing.to_value(self._unit)

            elif self.number == 0:
                return [] * self._unit, np.nan * self._unit

            elif self.number is not None:
                # number of ticks was specified, work out optimal spacing

                # first compute the exact spacing
                dv = abs(float(value_max - value_min)) / self.number * self._unit

                if self.format is not None and dv < self.base_spacing:
                    # if the spacing is less than the minimum spacing allowed by the format, simply
                    # use the format precision instead.
                    spacing_value = self.base_spacing.to_value(self._unit)
                else:
                    # otherwise we clip to the nearest 'sensible' spacing
                    if self.decimal:
                        from .utils import select_step_scalar

                        spacing_value = select_step_scalar(
                            dv.to_value(self._format_unit)
                        ) * self._format_unit.to(self._unit)
                    else:
                        if self._format_unit is u.degree:
                            from .utils import select_step_degree

                            spacing_value = select_step_degree(dv).to_value(self._unit)
                        else:
                            from .utils import select_step_hour

                            spacing_value = select_step_hour(dv).to_value(self._unit)

            # We now find the interval values as multiples of the spacing and
            # generate the tick positions from this.
            values = self._locate_values(value_min, value_max, spacing_value)
            return values * spacing_value * self._unit, spacing_value * self._unit

    def formatter(self, values, spacing, format="auto"):
        if not isinstance(values, u.Quantity) and values is not None:
            raise TypeError("values should be a Quantities array")

        if len(values) > 0:
            decimal = self.decimal
            unit = self._format_unit

            if unit is u.hour:
                unit = u.hourangle

            if self.format is None:
                if decimal:
                    # Here we assume the spacing can be arbitrary, so for example
                    # 1.000223 degrees, in which case we don't want to have a
                    # format that rounds to degrees. So we find the number of
                    # decimal places we get from representing the spacing as a
                    # string in the desired units. The easiest way to find
                    # the smallest number of decimal places required is to
                    # format the number as a decimal float and strip any zeros
                    # from the end. We do this rather than just trusting e.g.
                    # str() because str(15.) == 15.0. We format using 10 decimal
                    # places by default before stripping the zeros since this
                    # corresponds to a resolution of less than a microarcecond,
                    # which should be sufficient.
                    spacing = spacing.to_value(unit)
                    fields = 0
                    precision = len(
                        f"{spacing:.10f}".replace("0", " ").strip().split(".", 1)[1]
                    )
                else:
                    spacing = spacing.to_value(unit / 3600)
                    if spacing >= 3600:
                        fields = 1
                        precision = 0
                    elif spacing >= 60:
                        fields = 2
                        precision = 0
                    elif spacing >= 1:
                        fields = 3
                        precision = 0
                    else:
                        fields = 3
                        precision = -int(np.floor(np.log10(spacing)))
            else:
                fields = self._fields
                precision = self._precision

            is_latex = format == "latex" or (
                format == "auto" and rcParams["text.usetex"]
            )

            if decimal:
                if self.show_decimal_unit:
                    sep = "fromunit"
                    if is_latex:
                        fmt = "latex"
                    else:
                        if unit is u.hourangle:
                            fmt = "unicode"
                        else:
                            fmt = "generic"
                    unit = CUSTOM_UNITS.get(unit, unit)
                else:
                    sep = "fromunit"
                    fmt = None
            elif self.sep is not None:
                sep = self.sep
                fmt = None
            else:
                sep = "fromunit"
                if unit == u.degree:
                    if is_latex:
                        fmt = "latex"
                    else:
                        sep = ("\xb0", "'", '"')
                        fmt = None
                else:
                    if format == "ascii":
                        fmt = None
                    elif is_latex:
                        fmt = "latex"
                    else:
                        # Here we still use LaTeX but this is for Matplotlib's
                        # LaTeX engine - we can't use fmt='latex' as this
                        # doesn't produce LaTeX output that respects the fonts.
                        sep = (
                            r"$\mathregular{^h}$",
                            r"$\mathregular{^m}$",
                            r"$\mathregular{^s}$",
                        )
                        fmt = None

            angles = Angle(values)
            string = angles.to_string(
                unit=unit,
                precision=precision,
                decimal=decimal,
                fields=fields,
                sep=sep,
                format=fmt,
                alwayssign=self._alwayssign,
            ).tolist()

            return _fix_minus(string)
        else:
            return []


class ScalarFormatterLocator(BaseFormatterLocator):
    """
    A joint formatter/locator.
    """

    def __init__(
        self,
        values=None,
        number=None,
        spacing=None,
        format=None,
        unit=None,
        format_unit=None,
    ):
        if unit is None:
            if spacing is not None:
                unit = spacing.unit
            elif values is not None:
                unit = values.unit
        format_unit = format_unit or unit

        super().__init__(
            values=values,
            number=number,
            spacing=spacing,
            format=format,
            unit=unit,
            format_unit=format_unit,
        )

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, spacing):
        if spacing is not None and not isinstance(spacing, u.Quantity):
            raise TypeError("spacing should be an astropy.units.Quantity instance")
        self._number = None
        self._spacing = spacing
        self._values = None

    @property
    def format(self):
        return self._format

    @format.setter
    def format(self, value):
        self._format = value

        if value is None:
            return

        if SCAL_RE.match(value) is not None:
            if "." in value:
                self._precision = len(value) - value.index(".") - 1
            else:
                self._precision = 0

            if self.spacing is not None and self.spacing < self.base_spacing:
                warnings.warn(
                    "Spacing is too small - resetting spacing to match format"
                )
                self.spacing = self.base_spacing

            if self.spacing is not None:
                ratio = (self.spacing / self.base_spacing).decompose().value
                remainder = ratio - np.round(ratio)

                if abs(remainder) > 1.0e-10:
                    warnings.warn(
                        "Spacing is not a multiple of base spacing - resetting spacing"
                        " to match format"
                    )
                    self.spacing = self.base_spacing * max(1, round(ratio))

        elif not value.startswith("%"):
            raise ValueError(f"Invalid format: {value}")

    @property
    def base_spacing(self):
        return self._format_unit / (10.0**self._precision)

    def locator(self, value_min, value_max):
        if self.values is not None:
            # values were manually specified
            return self.values, 1.1 * self._unit
        else:
            # In the special case where value_min is the same as value_max, we
            # don't locate any ticks. This can occur for example when taking a
            # slice for a cube (along the dimension sliced).
            if value_min == value_max:
                return [] * self._unit, 0 * self._unit

            if self.spacing is not None:
                # spacing was manually specified
                spacing = self.spacing.to_value(self._unit)

            elif self.number is not None:
                # number of ticks was specified, work out optimal spacing

                # first compute the exact spacing
                dv = abs(float(value_max - value_min)) / self.number * self._unit

                if (
                    self.format is not None
                    and (not self.format.startswith("%"))
                    and dv < self.base_spacing
                ):
                    # if the spacing is less than the minimum spacing allowed by the format, simply
                    # use the format precision instead.
                    spacing = self.base_spacing.to_value(self._unit)
                else:
                    from .utils import select_step_scalar

                    spacing = select_step_scalar(
                        dv.to_value(self._format_unit)
                    ) * self._format_unit.to(self._unit)

            # We now find the interval values as multiples of the spacing and
            # generate the tick positions from this

            values = self._locate_values(value_min, value_max, spacing)
            return values * spacing * self._unit, spacing * self._unit

    def formatter(self, values, spacing, format="auto"):
        if len(values) > 0:
            if self.format is None:
                if spacing.value < 1.0:
                    precision = -int(np.floor(np.log10(spacing.value)))
                else:
                    precision = 0
            elif self.format.startswith("%"):
                return [(self.format % x.value) for x in values]
            else:
                precision = self._precision

            return _fix_minus(
                [
                    ("{0:." + str(precision) + "f}").format(
                        x.to_value(self._format_unit)
                    )
                    for x in values
                ]
            )

        else:
            return []



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Configure the tests for :mod:`astropy.cosmology`."""

from astropy.cosmology._src.tests.helper import clean_registry  # noqa: F401
from astropy.tests.helper import pickle_protocol  # noqa: F401



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import numpy as np

from astropy.coordinates.baseframe import frame_transform_graph
from astropy.coordinates.matrix_utilities import matrix_transpose
from astropy.coordinates.transformations import DynamicMatrixTransform

from .fk4 import FK4NoETerms
from .fk5 import FK5
from .utils import EQUINOX_B1950, EQUINOX_J2000

# FK5 to/from FK4 ------------------->
# B1950->J2000 matrix from Murray 1989 A&A 218,325 eqn 28
_B1950_TO_J2000_M = np.array(
    [
        [0.9999256794956877, -0.0111814832204662, -0.0048590038153592],
        [0.0111814832391717, +0.9999374848933135, -0.0000271625947142],
        [0.0048590037723143, -0.0000271702937440, +0.9999881946023742],
    ]
)

_FK4_CORR = (
    np.array(
        [
            [-0.0026455262, -1.1539918689, +2.1111346190],
            [+1.1540628161, -0.0129042997, +0.0236021478],
            [-2.1112979048, -0.0056024448, +0.0102587734],
        ]
    )
    * 1.0e-6
)


def _fk4_B_matrix(obstime):
    """
    This is a correction term in the FK4 transformations because FK4 is a
    rotating system - see Murray 89 eqn 29.
    """
    # Note this is *julian century*, not besselian
    T = (obstime.jyear - 1950.0) / 100.0
    if getattr(T, "shape", ()):
        # Ensure we broadcast possibly arrays of times properly.
        T.shape += (1, 1)
    return _B1950_TO_J2000_M + _FK4_CORR * T


# This transformation can't be static because the observation date is needed.
@frame_transform_graph.transform(DynamicMatrixTransform, FK4NoETerms, FK5)
def fk4_no_e_to_fk5(fk4noecoord, fk5frame):
    # Correction terms for FK4 being a rotating system
    B = _fk4_B_matrix(fk4noecoord.obstime)

    # construct both precession matrices - if the equinoxes are B1950 and
    # J2000, these are just identity matrices
    pmat1 = fk4noecoord._precession_matrix(fk4noecoord.equinox, EQUINOX_B1950)
    pmat2 = fk5frame._precession_matrix(EQUINOX_J2000, fk5frame.equinox)

    return pmat2 @ B @ pmat1


# This transformation can't be static because the observation date is needed.
@frame_transform_graph.transform(DynamicMatrixTransform, FK5, FK4NoETerms)
def fk5_to_fk4_no_e(fk5coord, fk4noeframe):
    # Get transposed version of the rotating correction terms... so with the
    # transpose this takes us from FK5/J200 to FK4/B1950
    B = matrix_transpose(_fk4_B_matrix(fk4noeframe.obstime))

    # construct both precession matrices - if the equinoxes are B1950 and
    # J2000, these are just identity matrices
    pmat1 = fk5coord._precession_matrix(fk5coord.equinox, EQUINOX_J2000)
    pmat2 = fk4noeframe._precession_matrix(EQUINOX_B1950, fk4noeframe.equinox)

    return pmat2 @ B @ pmat1



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The functions in this module provide faster versions of np.nan*
functions using the optional bottleneck package if it is installed. If
bottleneck is not installed, then the np.nan* functions are used.
"""

import functools
from collections.abc import Callable

import numpy as np
from numpy.typing import ArrayLike, NDArray

from astropy.stats.funcs import mad_std
from astropy.units import Quantity
from astropy.utils.compat.optional_deps import HAS_BOTTLENECK

if HAS_BOTTLENECK:
    import bottleneck

    def _move_tuple_axes_last(
        array: ArrayLike,
        axis: tuple[int, ...] | None = None,
    ) -> ArrayLike:
        """
        Move the specified axes of a NumPy array to the last positions
        and combine them.

        Bottleneck can only take integer axis, not tuple, so this
        function takes all the axes to be operated on and combines them
        into the last dimension of the array so that we can then use
        axis=-1.

        Parameters
        ----------
        array : `~numpy.ndarray`
            The input array.

        axis : tuple of int
            The axes on which to move and combine.

        Returns
        -------
        array_new : `~numpy.ndarray`
            Array with the axes being operated on moved into the last
            dimension.
        """
        other_axes = tuple(i for i in range(array.ndim) if i not in axis)

        # Move the specified axes to the last positions
        array_new = np.transpose(array, other_axes + axis)

        # Reshape the array by combining the moved axes
        return array_new.reshape(array_new.shape[: len(other_axes)] + (-1,))

    def _apply_bottleneck(
        function: Callable,
        array: ArrayLike,
        axis: int | tuple[int, ...] | None = None,
        **kwargs,
    ) -> float | NDArray | Quantity:
        """Wrap bottleneck function to handle tuple axis.

        Also takes care to ensure the output is of the expected type,
        i.e., a quantity, numpy array, or numpy scalar.
        """
        if isinstance(axis, tuple):
            array = _move_tuple_axes_last(array, axis=axis)
            axis = -1

        result = function(array, axis=axis, **kwargs)
        if isinstance(array, Quantity):
            if function is bottleneck.nanvar:
                result = array._result_as_quantity(result, array.unit**2, None)
            else:
                result = array.__array_wrap__(result)
            return result
        elif isinstance(result, float):
            # For compatibility with numpy, always return a numpy scalar.
            return np.float64(result)
        else:
            return result

    bn_funcs = dict(
        nansum=functools.partial(_apply_bottleneck, bottleneck.nansum),
        nanmin=functools.partial(_apply_bottleneck, bottleneck.nanmin),
        nanmax=functools.partial(_apply_bottleneck, bottleneck.nanmax),
        nanmean=functools.partial(_apply_bottleneck, bottleneck.nanmean),
        nanmedian=functools.partial(_apply_bottleneck, bottleneck.nanmedian),
        nanstd=functools.partial(_apply_bottleneck, bottleneck.nanstd),
        nanvar=functools.partial(_apply_bottleneck, bottleneck.nanvar),
    )

    np_funcs = dict(
        nansum=np.nansum,
        nanmin=np.nanmin,
        nanmax=np.nanmax,
        nanmean=np.nanmean,
        nanmedian=np.nanmedian,
        nanstd=np.nanstd,
        nanvar=np.nanvar,
    )

    def _dtype_dispatch(func_name):
        # dispatch to bottleneck or numpy depending on the input array dtype
        # this is done to workaround known accuracy bugs in bottleneck
        # affecting float32 calculations
        # see https://github.com/pydata/bottleneck/issues/379
        # see https://github.com/pydata/bottleneck/issues/462
        # see https://github.com/astropy/astropy/issues/17185
        # see https://github.com/astropy/astropy/issues/11492
        def wrapped(*args, **kwargs):
            if args[0].dtype.str[1:] == "f8":
                return bn_funcs[func_name](*args, **kwargs)
            else:
                return np_funcs[func_name](*args, **kwargs)

        return wrapped

    nansum = _dtype_dispatch("nansum")
    nanmin = _dtype_dispatch("nanmin")
    nanmax = _dtype_dispatch("nanmax")
    nanmean = _dtype_dispatch("nanmean")
    nanmedian = _dtype_dispatch("nanmedian")
    nanstd = _dtype_dispatch("nanstd")
    nanvar = _dtype_dispatch("nanvar")

else:
    nansum = np.nansum
    nanmin = np.nanmin
    nanmax = np.nanmax
    nanmean = np.nanmean
    nanmedian = np.nanmedian
    nanstd = np.nanstd
    nanvar = np.nanvar


def nanmadstd(
    array: ArrayLike,
    axis: int | tuple[int, ...] | None = None,
) -> float | NDArray:
    """mad_std function that ignores NaNs by default."""
    return mad_std(array, axis=axis, ignore_nan=True)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module tests some methods related to ``tdat`` format
reader/writer.
Requires `pyyaml <https://pyyaml.org/>`_ to be installed.
"""

import copy
import warnings
from collections import OrderedDict
from io import StringIO

import numpy as np
import pytest

from astropy import units as u
from astropy.io import ascii
from astropy.io.ascii.tdat import TdatFormatError, TdatFormatWarning
from astropy.table import QTable, Table
from astropy.table.table_helpers import simple_table

test_dat = [
    "<HEADER>",
    "#",
    "// TABLE: heasarc_simple",
    "# TOTAL ROWS: 7",
    "#",
    "table_name=heasarc_simple",
    'table_description = "Test table"',
    "table_security   =    public",
    "#",
    "# Table Parameters",
    "#",
    "field[record_number] = int4  [meta.id] (key) // Unique Identifier for Entry",
    "field[id] = int4  [meta.id] (index) // Source ID Number",
    "field[name] = char12  [meta.id;meta.main] (index) // String Name",
    "field[ra] = float8:.4f_degree [pos.eq.ra] (index) // Right Ascension",
    "field[dec] = float8:.4f_degree [pos.eq.dec] (index) // Declination",
    "field[empty] = float8:.4f // Empty // Comment",
    "#",
    "parameter_defaults = name ra dec",
    "#",
    "# Virtual Parameters",
    "#",
    "frequency_regime = Gamma-ray",
    "observatory_name = GAMMA-RAY BURSTS",
    "row_type = GRB",
    "table_author = Example et al.",
    "table_priority = 3.01",
    "table_type = Observation",
    "unique_key = record_number",
    "#",
    "# Data Format Specification",
    "#",
    "line[1] = record_number id name ra dec empty",
    "#",
    "<DATA>",
    "1|10|aaa|1.0|1.0||",
    "2|20|b|2.0|||",
    "3|30|c||3.0||",
    "4|20|||||",
    "5||||||",
    "|60|f|6.0|6.0||",
    "7| 70| g | 7.0 |7.0||",
    "<END>",
]
test_table = Table.read(test_dat, format="ascii.tdat")
# Corresponds to simple_table()
SIMPLE_LINES = [
    "<HEADER>",
    "table_name = astropy_table",
    "#",
    "# Table Parameters",
    "#",
    "field[a] = int4",
    "field[b] = float8",
    "field[c] = char1",
    "#",
    "# Data Format Specification",
    "#",
    "line[1] = a b c",
    "#",
    "<DATA>",
    "1|1.0|c|",
    "2|2.0|d|",
    "3|3.0|e|",
    "<END>",
]


def test_write_simple():
    """
    Write a simple table with common types.  This shows the compact version
    of serialization with one line per column.
    Table -> TDAT
    """
    t = simple_table()
    t.meta["table_name"] = "astropy_table"
    out = StringIO()
    t.write(out, format="ascii.tdat")

    # Explicit check that output matches expected
    assert out.getvalue().splitlines() == SIMPLE_LINES

    t.columns["a"][2] = -3
    t.columns["b"][2] = -3.0
    t.write(out, format="ascii.tdat")
    assert out.getvalue().splitlines()[-2] == "-3|-3.0|e|"


def test_catch_format():
    """
    Ensure that a table handles header keywords appropriately.
    Table -> TDAT
    """
    t = simple_table()
    out = StringIO()

    # Missing table_name
    with pytest.warns(TdatFormatWarning, match="'table_name' must be specified"):
        t.write(out, format="ascii.tdat")

    # table_name too long (>20 characters)
    with pytest.warns(TdatFormatWarning, match="'table_name' is too long"):
        t.meta["table_name"] = "origin_12345678901234567890"
        t.write(out, format="ascii.tdat")

        # check if the name is truncated
        nt = Table.read(out.getvalue(), format="ascii.tdat")
        assert nt.meta["keywords"]["table_name"] == t.meta["table_name"][:20]

    t.meta["table_name"] = "astropy_table"

    # table_description too long (>80 characters)
    with pytest.warns(TdatFormatWarning, match="'table_description' is too long"):
        t.meta["table_description"] = (
            """\
        This is a description that exceeds the character limit allowed by the tdat
        format and it should be truncated before being written. A warning should
        pop up to inform the user of this behavior.
        """.replace("\n", " ")
            .replace("\t", "")
            .strip()
        )
        t.write(out, format="ascii.tdat")

        # check if the description is truncated
        nt = Table.read(out.getvalue(), format="ascii.tdat")
        assert (
            nt.meta["keywords"]["table_description"]
            == t.meta["table_description"][:80].strip()
        )

    # table_security unrecognized value
    with pytest.warns(TdatFormatWarning, match="table_security not recognized"):
        t.meta["table_description"] = "Test Table"
        t.meta["table_security"] = "fortknox"
        t.write(out, format="ascii.tdat")


def test_read_tdat():
    """Ensure the Table is as it should be
    TDAT -> Table
    """
    assert test_table.meta["keywords"] == OrderedDict(
        [
            ("table_name", "heasarc_simple"),
            ("table_description", "Test table"),
            ("table_security", "public"),
            ("parameter_defaults", "name ra dec"),
            ("frequency_regime", "Gamma-ray"),
            ("observatory_name", "GAMMA-RAY BURSTS"),
            ("row_type", "GRB"),
            ("table_author", "Example et al."),
            ("table_priority", "3.01"),
            ("table_type", "Observation"),
            ("unique_key", "record_number"),
        ]
    )

    # Column checks
    descriptions = [
        "Unique Identifier for Entry",
        "Source ID Number",
        "String Name",
        "Right Ascension",
        "Declination",
        "Empty",
    ]
    dtypes = [np.int32, np.int32, np.dtype("U3"), float, float, float]
    units = [None, None, None, "deg", "deg", None]
    meta = [
        {"ucd": "meta.id", "index": "key"},
        {"ucd": "meta.id", "index": "index"},
        {"ucd": "meta.id;meta.main", "index": "index"},
        {"ucd": "pos.eq.ra", "index": "index"},
        {"ucd": "pos.eq.dec", "index": "index"},
        {"comment": "Comment"},
    ]
    for i, col in enumerate(test_table.itercols()):
        assert col.description == descriptions[i]
        assert col.dtype == dtypes[i]
        assert col.unit == units[i]
        assert col.meta == meta[i]

    # data check
    ## Missing is masked
    assert isinstance(test_table["ra"][3], np.ma.core.MaskedConstant)
    ## Check data matches simple csv format
    test_data = [
        "record_number, id, name, ra, dec, empty",
        "1, 10, aaa, 1.0, 1.0, ",
        "2, 20,   b, 2.0,    , ",
        "3, 30,   c,    , 3.0, ",
        "4, 20,    ,    ,    , ",
        "5,   ,    ,    ,    , ",
        " , 60,   f, 6.0, 6.0, ",
        "7, 70,   g, 7.0, 7.0, ",
    ]
    table_data = Table.read(test_data, format="csv")
    assert all((table_data == test_table).data)


def test_full_table_content():
    """Check the table content matches expectation.
    TDAT -> Table
    """
    assert test_table.meta == {
        "comments": ["TABLE: heasarc_simple", "TOTAL ROWS: 7"],
        "keywords": OrderedDict(
            [
                ("table_name", "heasarc_simple"),
                ("table_description", "Test table"),
                ("table_security", "public"),
                ("parameter_defaults", "name ra dec"),
                ("frequency_regime", "Gamma-ray"),
                ("observatory_name", "GAMMA-RAY BURSTS"),
                ("row_type", "GRB"),
                ("table_author", "Example et al."),
                ("table_priority", "3.01"),
                ("table_type", "Observation"),
                ("unique_key", "record_number"),
            ]
        ),
    }
    assert len(test_table) == 7
    dtypes = [np.int32, np.int32, np.dtype("U3"), float, float, float]
    descriptions = [
        "Unique Identifier for Entry",
        "Source ID Number",
        "String Name",
        "Right Ascension",
        "Declination",
        "Empty",
    ]
    ucds = ["meta.id", "meta.id", "meta.id;meta.main", "pos.eq.ra", "pos.eq.dec", None]
    indices = ["key", "index", "index", "index", "index", None]
    for i, col in enumerate(test_table.columns):
        assert test_table[col].dtype == dtypes[i]
        assert test_table[col].description == descriptions[i]
        assert test_table[col].meta.get("ucd", None) == ucds[i]
        assert test_table[col].meta.get("index", None) == indices[i]
    assert test_table["ra"].unit == "deg"
    assert test_table["dec"].unit == "deg"
    assert test_table["ra"].format == ".4f"
    assert test_table["dec"].format == ".4f"
    assert test_table["empty"].format == ".4f"


def test_write_full():
    """
    Write a full-featured table with common types and explicitly check output
    TDAT -> Table -> TDAT (formatted)

    Differences between `lines` and `test_dat`:
    - Empty comment lines are dropped (except when demarcating a section header
    like "Table Parameters").
    - The data type for string field "name" is downsized to char3 from char12
    reflecting the actual maximum size in the column.
    - Extraneous spaces in the data are stripped
    These differences reflect flexibility in reading in from a tdat file, and
    writing out in a standardized way.
    """
    t = test_table

    lines = [
        "<HEADER>",
        "# TABLE: heasarc_simple",
        "# TOTAL ROWS: 7",
        "table_name = heasarc_simple",
        "table_description = Test table",
        "table_security = public",
        "#",
        "# Table Parameters",
        "#",
        "field[record_number] = int4 [meta.id] (key) // Unique Identifier for Entry",
        "field[id] = int4 [meta.id] (index) // Source ID Number",
        "field[name] = char3 [meta.id;meta.main] (index) // String Name",
        "field[ra] = float8:.4f_deg [pos.eq.ra] (index) // Right Ascension",
        "field[dec] = float8:.4f_deg [pos.eq.dec] (index) // Declination",
        "field[empty] = float8:.4f // Empty // Comment",
        "#",
        "parameter_defaults = name ra dec",
        "#",
        "# Virtual Parameters",
        "#",
        "frequency_regime = Gamma-ray",
        "observatory_name = GAMMA-RAY BURSTS",
        "row_type = GRB",
        "table_author = Example et al.",
        "table_priority = 3.01",
        "table_type = Observation",
        "unique_key = record_number",
        "#",
        "# Data Format Specification",
        "#",
        "line[1] = record_number id name ra dec empty",
        "#",
        "<DATA>",
        "1|10|aaa|1.0000|1.0000||",
        "2|20|b|2.0000|||",
        "3|30|c||3.0000||",
        "4|20|||||",
        "5||||||",
        "|60|f|6.0000|6.0000||",
        "7|70|g|7.0000|7.0000||",
        "<END>",
    ]

    out = StringIO()
    t.write(out, format="ascii.tdat")
    assert out.getvalue().splitlines() == lines


def test_write_read_roundtrip():
    """
    Write a full-featured table with all types and see that it round-trips on
    readback.
    TDAT -> Table -> TDAT (formatted) -> Table
    Tables should have the same content.
    """
    t = test_table
    out = StringIO()
    t.write(out, format="ascii.tdat")

    t2s = [
        Table.read(out.getvalue(), format="ascii.tdat"),
        ascii.read(out.getvalue(), format="tdat"),
    ]
    for t2 in t2s:
        assert t.meta == t2.meta
        for name in t.colnames:
            assert t[name].attrs_equal(t2[name])
            assert np.all(t[name] == t2[name]) or np.all(t[name].mask == t2[name].mask)


def test_write_read_roundtrip_empty_table(tmp_path):
    """see https://github.com/astropy/astropy/issues/13191"""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=TdatFormatWarning)
        sfile = tmp_path / "x.tdat"
        Table().write(sfile)
        t = Table.read(sfile)
        assert len(t) == 0
        assert len(t.colnames) == 0


def test_keyword_quotes():
    lines = copy.copy(SIMPLE_LINES)
    # double quotes
    lines[1] = 'table_name = "astropy_table"'
    t = Table.read(lines, format="ascii.tdat")
    assert t.meta["keywords"]["table_name"] == "astropy_table"
    # single quotes
    lines[1] = "table_name = 'astropy_table'"
    t = Table.read(lines, format="ascii.tdat")
    assert t.meta["keywords"]["table_name"] == "astropy_table"
    # back quotes
    lines[1] = "table_name = `astropy_table`"
    t = Table.read(lines, format="ascii.tdat")
    assert t.meta["keywords"]["table_name"] == "astropy_table"
    # combination and multiple, nested properly
    lines[1] = "table_name = \"'`astropy_table`'\""
    t = Table.read(lines, format="ascii.tdat")
    assert t.meta["keywords"]["table_name"] == "astropy_table"

    # mismatched
    lines[1] = "table_name = \"astropy_table'"
    with pytest.raises(TdatFormatError, match="Mismatched"):
        t = Table.read(lines, format="ascii.tdat")
    # combination, nested improperly
    lines[1] = "table_name = \"'astropy_table\"'"
    with pytest.raises(TdatFormatError, match="Mismatched"):
        t = Table.read(lines, format="ascii.tdat")


def test_tablenames():
    """_summary_"""
    lines = copy.copy(SIMPLE_LINES)
    t = Table.read(lines, format="ascii.tdat")
    out = StringIO()

    with pytest.warns(TdatFormatWarning, match="'table_name' is too long"):
        t.meta["table_name"] = "origin_12345678901234567890"
        t.write(out, format="ascii.tdat")


def test_bad_delimiter():
    """
    Passing a delimiter other than | (pipe) gives an exception
    """
    out = StringIO()

    with pytest.warns(
        TdatFormatWarning, match="Delimiters other than the pipe character"
    ):
        test_table.write(out, format="ascii.tdat", delimiter=",")

    check_table = test_table.copy()
    check_table.meta["keywords"]["field_delimiter"] = ","
    with pytest.warns(
        TdatFormatWarning, match="Delimiters other than the pipe character"
    ):
        with pytest.warns(TdatFormatWarning, match="Skipping deprecated"):
            check_table.write(out, format="ascii.tdat", delimiter=",")


def test_bad_header_start():
    """
    Bad header without initial <HEADER>
    """
    lines = copy.copy(SIMPLE_LINES)
    lines[0] = "<DATA>"
    with pytest.raises(TdatFormatError) as err:
        Table.read("\n".join(lines), format="ascii.tdat")
        assert "<HEADER> not found in file." in str(err.value)


def test_missing_data_heading():
    """No data heading delimiter"""
    lines = copy.copy(SIMPLE_LINES)
    # Mislabelled
    lines[13] = "<DAYA>"
    with pytest.raises(TdatFormatError, match="<DATA> not found in file"):
        Table.read("\n".join(lines), format="ascii.tdat")
    # Not present at all
    del lines[13]
    with pytest.raises(TdatFormatError, match="<DATA> not found in file"):
        Table.read("\n".join(lines), format="ascii.tdat")


def test_bad_end_heading():
    """
    No <END> heading - should not raise an error
    """
    lines = copy.copy(SIMPLE_LINES)
    del lines[-1]
    Table.read("\n".join(lines), format="ascii.tdat")


def test_unrecognized_dtype():
    """Not all dtypes are supported by tdat files"""
    lines = copy.copy(SIMPLE_LINES)
    lines[5] = "field[a] = complex"
    with pytest.raises(TdatFormatError, match="unsupported data type"):
        Table.read("\n".join(lines), format="ascii.tdat")


def test_supported_dtypes():
    """Ensure all supported dtypes import and export as expected"""
    dtype_dict = {
        "int1": "int1",
        "integer1": "int1",
        "tinyint": "int1",
        "int2": "int2",
        "integer2": "int2",
        "smallint": "int2",
        "int4": "int4",
        "integer4": "int4",
        "integer": "int4",
        "float4": "float4",
        "real": "float4",
        "float": "float8",
        "float8": "float8",
    }
    lines = copy.copy(SIMPLE_LINES)

    # integer dtypes
    int_dtypes = [
        "int1",
        "integer1",
        "tinyint",
        "int2",
        "integer2",
        "smallint",
        "int4",
        "integer4",
        "integer",
    ]
    for dtype in int_dtypes:
        lines[5] = f"field[a]  = {dtype}"
        t = Table.read("\n".join(lines), format="ascii.tdat")
        assert all(t["a"].data == np.array([1, 2, 3]))
        out = StringIO()
        t.write(out, format="ascii.tdat")
        assert dtype_dict[dtype] in out.getvalue().splitlines()[5]

    # float dtypes
    float_dtypes = ["float4", "real", "float", "float8"]
    for dtype in float_dtypes:
        lines[6] = f"field[b]  = {dtype}"
        t = Table.read("\n".join(lines), format="ascii.tdat")
        assert all(t["b"].data == np.array([1.0, 2.0, 3.0]))
        out = StringIO()
        t.write(out, format="ascii.tdat")
        assert dtype_dict[dtype] in out.getvalue().splitlines()[6]

    # string dtypes
    for line in ["field[c] = char(3)", "field[c] = char(3) (index)"]:
        lines[7] = line
        t = Table.read("\n".join(lines), format="ascii.tdat")
        assert all(t["c"].data == np.array(["c", "d", "e"]))
        out = StringIO()
        t.write(out, format="ascii.tdat")
        assert "char1" in out.getvalue().splitlines()[7]  # downsizes to min length


def test_mismatch_line_field():
    """Not all dtypes are supported by tdat files"""
    lines = copy.copy(SIMPLE_LINES)
    lines[11] = "line[1] = a b c d"
    with pytest.raises(TdatFormatError, match='columns "field" descriptors'):
        Table.read("\n".join(lines), format="ascii.tdat")


def test_fieldname_too_long():
    """Test if field name being too long throws the expected error"""
    lines = copy.copy(SIMPLE_LINES)
    lines[5] = f"field[{'a' * 30}] = int4"
    lines[11] = f"line[1] = {'a' * 30} b c"
    with pytest.warns(TdatFormatWarning, match="field name must be shorter"):
        t = Table.read(lines, format="ascii.tdat")

    with pytest.warns(TdatFormatWarning, match="truncating."):
        out = StringIO()
        t.write(out, format="ascii.tdat")

    assert out.getvalue().splitlines()[5] == f"field[{'a' * 23}] = int4"


def test_fmt_type_too_long():
    """The combination of type and format has a maximum character length of 24"""
    lines = copy.copy(SIMPLE_LINES)
    lines[6] = "field[b] = float8:.0000000000000000000000000000001f"
    with pytest.raises(TdatFormatError, match="The type:fmt specifier"):
        Table.read("\n".join(lines), format="ascii.tdat")


def test_units():
    """Test that units act as expected (astropy converts some units)"""
    lines = copy.copy(SIMPLE_LINES)
    lines[6] = "field[b] = float8:.4f_m/s"
    t = Table.read(lines, format="ascii.tdat")
    # Astropy automatically interprets units
    assert t.columns["b"].unit == u.m / u.s

    out = StringIO()
    t.write(out, format="ascii.tdat")
    # When writing, astropy units will be written out using the cds format
    # so that there are no spaces
    assert "m.s-1" in out.getvalue().splitlines()[6]
    t2 = Table.read(out.getvalue(), format="ascii.tdat")
    assert t2.columns["b"].unit == t.columns["b"].unit

    lines[6] = "field[b] = float8:.4f_fake1/fake2"
    t = Table.read(lines, format="ascii.tdat")
    # Astropy automatically interprets units
    assert isinstance(t.columns["b"].unit, u.UnrecognizedUnit)
    assert str(t.columns["b"].unit) == "fake1/fake2"
    out = StringIO()
    t.write(out, format="ascii.tdat")
    # unrecognized units are not reformatted
    assert "fake1/fake2" in out.getvalue().splitlines()[6]


def assert_objects_equal(obj1, obj2, attrs, compare_class=True):
    if compare_class:
        assert obj1.__class__ is obj2.__class__

    assert obj1.shape == obj2.shape

    info_attrs = [
        "info.name",
        "info.format",
        "info.unit",
        "info.description",
        "info.dtype",
    ]
    for attr in attrs + info_attrs:
        a1 = obj1
        a2 = obj2
        for subattr in attr.split("."):
            try:
                a1 = getattr(a1, subattr)
                a2 = getattr(a2, subattr)
            except AttributeError:
                a1 = a1[subattr]
                a2 = a2[subattr]

        if isinstance(a1, np.ndarray) and a1.dtype.kind == "f":
            assert u.quantity_allclose(a1, a2, rtol=1e-10)
        else:
            assert np.all(a1 == a2)

    # For no attrs that means we just compare directly.
    if not attrs:
        if isinstance(obj1, np.ndarray) and obj1.dtype.kind == "f":
            assert u.quantity_allclose(obj1, obj2, rtol=1e-15)
        else:
            assert np.all(obj1 == obj2)


def test_round_trip_masked_table_default(tmp_path):
    """Test (mostly) round-trip of MaskedColumn through tdat using default serialization
    that uses an empty string "" to mark NULL values.  Note:

    >>> simple_table(masked=True)
    <Table masked=True length=3>
      a      b     c
    int64 float64 str1
    ----- ------- ----
       --     1.0    c
        2     2.0   --
        3      --    e
    """
    filename = tmp_path / "x.tdat"

    t = simple_table(masked=True)  # int, float, and str cols with one masked element
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=TdatFormatWarning)
        t.write(filename)

        t2 = Table.read(filename)
        assert t2.masked is False
        assert t2.colnames == t.colnames
        for name in t2.colnames:
            # From formal perspective the round-trip columns are the "same"
            assert np.all(t2[name].mask == t[name].mask)
            assert np.all(t2[name] == t[name])

            # But peeking under the mask shows that the underlying data are changed
            # because by default ECSV uses "" to represent masked elements.
            t[name].mask = False
            t2[name].mask = False
            assert not np.all(t2[name] == t[name])  # Expected diff


def test_deprecated_keyword():
    """Deprecated and obsolete keywords should raise warnings"""
    test_data = test_dat.copy()
    data_lines = test_data[35:-1]
    data_lines = "\123".join(data_lines)  # now a single string with a delimiter
    data_lines = data_lines.replace("|", "!")
    del test_data[35:-1]
    test_data.insert(35, data_lines)
    test_data.insert(8, 'record_delimiter = "\123"')
    test_data.insert(8, 'field_delimiter = "!"')
    with pytest.warns(TdatFormatWarning, match="keyword is deprecated"):
        t = Table.read(test_data, format="ascii.tdat")

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=TdatFormatWarning)
        t = Table.read(test_data, format="ascii.tdat")
        assert all(t == test_table)

    test_data = SIMPLE_LINES.copy()
    for i in range(14, 17):
        test_data[i] = test_data[i].replace("|", " ")
    test_data.insert(2, 'field_delimiter = " "')

    with pytest.warns(TdatFormatWarning, match="keyword is deprecated"):
        with pytest.warns(TdatFormatWarning, match="Double check your data"):
            t = Table.read(test_data, format="ascii.tdat")
            t2 = Table.read(SIMPLE_LINES, format="ascii.tdat")
            assert all(t == t2)

    test_data = test_dat.copy()
    test_data.insert(8, "relate[ra] = dec")
    with pytest.warns(TdatFormatWarning, match="keyword is obsolete"):
        t = Table.read(test_data, format="ascii.tdat")


def test_tdat_format_error():
    """Test the basic error message"""
    _STD_MSG = (
        "See details in https://heasarc.gsfc.nasa.gov/docs/software/dbdocs/tdat.html"
    )
    error_msg = "Test error message"
    with pytest.raises(TdatFormatError) as exc_info:
        raise TdatFormatError(error_msg)

    assert str(exc_info.value).startswith(error_msg)

    # Test that the standard message is appended
    assert _STD_MSG in str(exc_info.value)

    # Test with empty error message
    with pytest.raises(TdatFormatError) as exc_info:
        raise TdatFormatError()

    assert str(exc_info.value).startswith("\nSee details")

    # Test with a different error message
    another_msg = "Another test message"
    with pytest.raises(TdatFormatError) as exc_info:
        raise TdatFormatError(another_msg)

    assert str(exc_info.value).startswith(another_msg)
    assert _STD_MSG in str(exc_info.value)


def test_delimiter_in_data():
    """Test escaped delimiters in data"""
    lines = copy.deepcopy(test_dat)
    lines[38] = "4|20|\\|||||"
    t = Table.read(lines, format="ascii.tdat")
    assert t["name"][3] == "|"

    lines.insert(8, 'field_delimiter = "|$"')
    lines[39] = "4$20|\\|\\$||$|"
    with pytest.warns(TdatFormatWarning, match="keyword is deprecated"):
        t = Table.read(lines, format="ascii.tdat")
        assert t["name"][3] == "|$"
        out = StringIO()
        t.write(out, format="ascii.tdat")
        # Windows uses carriage return (\r) and line feed (\n).
        outval = out.getvalue().split("\n")[36].replace("\r", "")
        assert outval == "4|20|\\|$||||"


def test_unsupported_delimiter():
    """Test escaped delimiters in data"""
    lines = SIMPLE_LINES.copy()
    lines.insert(2, 'record_delimiter = "|"')
    with pytest.warns(TdatFormatWarning, match="keyword is deprecated"):
        with pytest.raises(TdatFormatError, match="'|' character reserved"):
            t = Table.read(lines, format="ascii.tdat")

        with pytest.raises(TdatFormatError, match="Unsupported record delimiter"):
            lines[2] = f'record_delimiter = "{chr(129)}"'
            t = Table.read(lines, format="ascii.tdat")

        with pytest.raises(TdatFormatError, match="Unsupported field delimiter"):
            lines[2] = f'field_delimiter = "{chr(129)}"'
            t = Table.read(lines, format="ascii.tdat")


def test_write_qtable():
    qt = QTable([np.arange(4) * u.m, ["a", "b", "c", "ddd"]], names=["a", "b"])
    qt.meta["table_name"] = "astropy_qtable"
    out = StringIO()
    qt.write(out, format="ascii.tdat")


def test_multiline_data():
    SIMPLE_LINES_MULTI = [
        "<HEADER>",
        "table_name = astropy_table",
        "#",
        "# Table Parameters",
        "#",
        "field[a] = int4",
        "field[b] = float8",
        "field[c] = char1",
        "#",
        "# Data Format Specification",
        "#",
        "line[1] = a b",
        "line[2] = c#",
        "<DATA>",
        "1|1.0|",
        "c|",
        "2|2.0|",
        "d|",
        "3|3.0|",
        "e|",
        "<END>",
    ]
    t1 = Table.read(SIMPLE_LINES, format="ascii.tdat")
    t2 = Table.read(SIMPLE_LINES_MULTI, format="ascii.tdat")
    assert all(t1 == t2)


def test_comment_too_long():
    lines = SIMPLE_LINES.copy()
    lines[5] += "# # " + "a" * 100
    with pytest.warns(TdatFormatWarning, match="Comments are limited"):
        t = Table.read(lines, format="ascii.tdat")
        assert len(t["a"].meta["comment"]) == 80


def test_int_too_large():
    t = simple_table()
    t.meta["table_name"] = "astropy_table"
    out = StringIO()
    # within limits
    t["a"][0] = 2**31 - 1
    t.write(out, format="ascii.tdat")
    t["a"][0] = -(2**31)
    t.write(out, format="ascii.tdat")
    # too large
    t["a"][0] = 2**31
    with pytest.raises(TdatFormatError, match="cannot be converted"):
        t.write(out, format="ascii.tdat")
    t["a"][0] = -(2**31) - 1
    with pytest.raises(TdatFormatError, match="cannot be converted"):
        t.write(out, format="ascii.tdat")



__all__ = ["DarkMatterComponent"]

from collections.abc import Callable
from typing import Any

from numpy.typing import ArrayLike, NDArray

from astropy.cosmology._src.typing import FArray
from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity


class DarkMatterComponent:
    """The cosmology has attributes and methods for the dark matter density.

    This trait provides an ``Odm`` method that returns the dark matter
    density parameter (i.e., total matter minus baryons) at redshift ``z``.
    """

    Odm0: float
    """Omega dark matter: dark matter density/critical density at z=0."""

    inv_efunc: Callable[[NDArray[Any]], NDArray[Any]]

    @deprecated_keywords("z", since="7.0")
    def Odm(self, z: Quantity | ArrayLike) -> FArray:
        """Return the density parameter for dark matter at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Odm : ndarray
            The density of dark matter relative to the critical density at
            each redshift.

        """
        z = aszarr(z)
        return self.Odm0 * (z + 1.0) ** 3 * self.inv_efunc(z) ** 2
